@@ -29,6 +29,7 @@ import java.net.InetAddress;
 import java.net.URL;
 import java.net.URLConnection;
 import java.util.HashMap;
+import java.util.LinkedHashMap;
 import java.util.Map;
 import java.util.Set;
 
@@ -50,7 +51,7 @@ public class GatewayDevice {
 	/**
 	 * Receive timeout when requesting data from device
 	 */
-    private static final int HTTP_RECEIVE_TIMEOUT = 7000;
+    private static final int DEFAULT_HTTP_RECEIVE_TIMEOUT = 7000;
     
 	private String st;
     private String location;
@@ -103,6 +104,11 @@ public class GatewayDevice {
      */
     private String modelName;
 
+    /**
+     * Timeout in milliseconds for HTTP reads
+     */
+    private static int httpReadTimeout = DEFAULT_HTTP_RECEIVE_TIMEOUT;
+
     /**
      * Creates a new instance of GatewayDevice
      */
@@ -123,7 +129,7 @@ public class GatewayDevice {
     public void loadDescription() throws SAXException, IOException {
 
         URLConnection urlConn = new URL(getLocation()).openConnection();
-        urlConn.setReadTimeout(HTTP_RECEIVE_TIMEOUT);
+        urlConn.setReadTimeout(httpReadTimeout);
 
         XMLReader parser = XMLReaderFactory.createXMLReader();
         parser.setContentHandler(new GatewayDeviceHandler(this));
@@ -197,7 +203,8 @@ public class GatewayDevice {
         HttpURLConnection conn = (HttpURLConnection) postUrl.openConnection();
 
         conn.setRequestMethod("POST");
-        conn.setReadTimeout(HTTP_RECEIVE_TIMEOUT);
+        conn.setConnectTimeout(httpReadTimeout);
+        conn.setReadTimeout(httpReadTimeout);
         conn.setDoOutput(true);
         conn.setRequestProperty("Content-Type", "text/xml");
         conn.setRequestProperty("SOAPAction", soapAction);
@@ -291,7 +298,7 @@ public class GatewayDevice {
     public boolean addPortMapping(int externalPort, int internalPort,
                                   String internalClient, String protocol, String description)
             throws IOException, SAXException {
-        Map<String, String> args = new HashMap<String, String>();
+        Map<String, String> args = new LinkedHashMap<String, String>();
         args.put("NewRemoteHost", "");    // wildcard, any remote host matches
         args.put("NewExternalPort", Integer.toString(externalPort));
         args.put("NewProtocol", protocol);
@@ -334,7 +341,7 @@ public class GatewayDevice {
         portMappingEntry.setExternalPort(externalPort);
         portMappingEntry.setProtocol(protocol);
 
-        Map<String, String> args = new HashMap<String, String>();
+        Map<String, String> args = new LinkedHashMap<String, String>();
         args.put("NewRemoteHost", ""); // wildcard, any remote host matches
         args.put("NewExternalPort", Integer.toString(externalPort));
         args.put("NewProtocol", protocol);
@@ -349,7 +356,7 @@ public class GatewayDevice {
                 !nameValue.containsKey("NewInternalPort"))
             return false;
 
-        portMappingEntry.setProtocol(nameValue.get("NewProtocol"));
+        portMappingEntry.setProtocol(protocol);
         portMappingEntry.setEnabled(nameValue.get("NewEnabled"));
         portMappingEntry.setInternalClient(nameValue.get("NewInternalClient"));
         portMappingEntry.setExternalPort(externalPort);
@@ -384,7 +391,7 @@ public class GatewayDevice {
     public boolean getGenericPortMappingEntry(int index,
                                               final PortMappingEntry portMappingEntry)
             throws IOException, SAXException {
-        Map<String, String> args = new HashMap<String, String>();
+        Map<String, String> args = new LinkedHashMap<String, String>();
         args.put("NewPortMappingIndex", Integer.toString(index));
 
         Map<String, String> nameValue = simpleUPnPcommand(controlURL,
@@ -451,7 +458,7 @@ public class GatewayDevice {
      */
     public boolean deletePortMapping(int externalPort, String protocol)
             throws IOException, SAXException {
-        Map<String, String> args = new HashMap<String, String>();
+        Map<String, String> args = new LinkedHashMap<String, String>();
         args.put("NewRemoteHost", "");
         args.put("NewExternalPort", Integer.toString(externalPort));
         args.put("NewProtocol", protocol);
@@ -633,6 +640,22 @@ public class GatewayDevice {
         this.modelNumber = modelNumber;
     }
 
+    /**
+     * Gets the timeout for actions on the device.
+     * @return timeout in milliseconds
+     */
+    public static int getHttpReadTimeout() {
+        return httpReadTimeout;
+    }
+
+    /**
+     * Sets the timeout for actions on the device.
+     * @param milliseconds the new timeout in milliseconds
+     */
+    public static void setHttpReadTimeout(int milliseconds) {
+        httpReadTimeout = milliseconds;
+    }
+
     // private methods
     private String copyOrCatUrl(String dst, String src) {
         if (src != null) {
